from typing import Tuple, Iterable
from copy import deepcopy
from abc import ABCMeta, abstractmethod
import numpy as np


class WeakReadonlyList(list):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)

    def __setitem__(self, key, value):
        raise RuntimeError("readonly list")

    def __delitem__(self, key):
        raise RuntimeError("readonly list")

    def append(self, value):
        raise RuntimeError("readonly list")

    def extend(self, values):
        raise RuntimeError("readonly list")

    def insert(self, index, value):
        raise RuntimeError("readonly list")

    def pop(self, index):
        raise RuntimeError("readonly list")

    def remove(self, value):
        raise RuntimeError("readonly list")

    def clear(self) -> None:
        raise RuntimeError("readonly list")

    def system_append(self, value):
        super().append(value)

    def system_extend(self, values):
        super().extend(values)

    def system_insert(self, index, value):
        super().insert(index, value)

    def system_pop(self, index):
        return super().pop(index)

    def system_remove(self, value):
        super().remove(value)

    def system_setitem(self, key, value):
        super().__setitem__(key, value)

    def system_delitem(self, key):
        super().__delitem__(key)

    def system_clear(self):
        super().clear()

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def __iadd__(self, other):
        raise RuntimeError("readonly list")


__cls_types = []


class DataIO(metaclass=ABCMeta):
    @abstractmethod
    def label2output(self, label: object) -> np.ndarray:
        """
        trans label to output
        :param label: object, general in float or str
        :return: output: ndarray, shape = model.output_shape[1:]
        """
        pass

    @abstractmethod
    def output2label(self, output: np.ndarray) -> object:
        """
        trans output to label
        :param output: ndarray, shape = model.output_shape[1:]
        :return: labels: object, general in float or str
        """
        pass


class DataPool(DataIO, metaclass=ABCMeta):
    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls)
        inst._bufx = []
        inst._bufy = []
        return inst

    @property
    def pool(self) -> Tuple[WeakReadonlyList, WeakReadonlyList]:
        return self._pool

    @property
    def x(self) -> WeakReadonlyList:
        return self._pool[0]

    @property
    def xshape(self) -> tuple:
        if len(self._pool[0]) == 0:
            raise ValueError("No any data in pool. Can not get the shape.")
        return np.array(self._pool[0][0]).shape

    @property
    def y(self) -> WeakReadonlyList:
        return self._pool[1]

    @property
    def yshape(self) -> tuple:
        if len(self._pool[1]) == 0:
            raise ValueError("No any data in pool. Can not get the shape.")
        return np.array(self._pool[1][0]).shape

    def __init__(self):
        self._pool: Tuple[WeakReadonlyList, WeakReadonlyList] = WeakReadonlyList(), WeakReadonlyList()  # x_list, y_list
        self._left = self.load(None)
        self._flag = False  # 是否已使用 FLAG | 一般只有第一次有效

    def __getitem__(self, item):
        if isinstance(item, str):
            if item.lower() == 'x':
                return self._pool[0]
            elif item.lower() == 'y':
                return self._pool[1]
            else:
                raise ValueError(f"'{item}' is not a valid str item")
        else:
            return self._pool[0][item], self._pool[1][item]

    def __len__(self):
        return len(self._pool[0])

    def append(self, x: np.ndarray, y: object):
        self._pool[0].system_append(x)
        self._pool[1].system_append(y)

    def extend(self, x: list, y: list):
        assert len(x) == len(y), f"x and y must have the same length, but x={len(x)}, y={len(y)}"
        self._pool[0].system_extend(x)
        self._pool[1].system_extend(y)

    def clear(self):
        self._pool[0].system_clear()
        self._pool[1].system_clear()

    @abstractmethod
    def load(self, left=None) -> None:
        pass

    @abstractmethod
    def augments(self, xs, ys) -> tuple:
        """
        对xs ys进行数据增强
        发生在call的第2. 进行数据增强
        :param xs:
        :param ys:
        :return:
        """
        pass

    @abstractmethod
    def prehandle(self, x) -> np.ndarray:
        pass

    @abstractmethod
    def posthandle(self, y) -> object:
        pass

    def __call__(self, num:int=None, percent:float=None):
        if num is None:
            if percent is None:
                raise ValueError('num or percent must be specified')

            assert (0 - 1e-6) <= percent <= (1 + 1e-6), 'percent must be in [0, 1]'
            num = int(len(self) * percent)
            if num <= 0:
                raise ValueError(f'Not enough data to train. at persent={percent}')


        while num > len(self._bufx):
            if self._flag and self._left is not None:
                self._left = self.load(self._left)
            self._bufx.extend(self._pool[0].copy())
            self._bufy.extend(self._pool[1].copy())
            self._flag = True

        _x, self._bufx = self._bufx[:num], self._bufx[num:]
        _y, self._bufy = self._bufy[:num], self._bufy[num:]

        _x, _y = self.augments(_x, _y)

        for i in range(len(_x)):
            _x[i] = self.prehandle(_x[i])
        for i in range(len(_y)):
            _y[i] = self.label2output(_y[i])
        return _x, _y


class DataControl(DataPool, metaclass=ABCMeta):
    @abstractmethod
    def label2output(self, label: object) -> np.ndarray:
        """
        trans label to output
        :param label: object, general in float or str
        :return: output: ndarray, shape = model.output_shape[1:]
        """
        pass

    @abstractmethod
    def output2label(self, output: np.ndarray) -> object:
        """
        trans output to label
        :param output: ndarray, shape = model.output_shape[1:]
        :return: labels: object, general in float or str
        """
        pass

    @abstractmethod
    def load(self, left=None) -> None:
        """
        load data from somewhere into self

        self.append(x: np.ndarray, y: object)
        self.x or self['x'] to get the WeakReadonlyList
        self.y or self['y'] to get the WeakReadonlyList
        self[?] to get the x[?], y[?]

        :param left: 上次调用load时, load的返回值, 应该是float. 第一次调用时传入None
        :return: left, 剩余量, 是一个float, 0 - 1, 如果一次就读完了, 返回None
        """
        pass

    # --------------------------------------------------------------------------
    # 下面是可选的方法:
    def augments(self, xs, ys) -> tuple:
        """
        对xs ys进行数据增强
        发生在call的第2. 进行数据增强
        :param xs:
        :param ys:
        :return:
        """
        return xs, ys


    def prehandle(self, x) -> np.ndarray:
        """
        对x进行预处理
        原理:
        预处理只在__call__以及LModel.predict中被调用, 其余方式获取x y都是原始的数据

        __call__的调用顺序:
        1. 取出原始数据集中的一部分.
        2. 进行数据增强
        3. 对每个数据的x进行预处理             <----------------------------------
        4. 将每个数据的y(label)转换为y(stdandard output)
        -> 最后返回 List[type(x)], List[type(y_output)]

        :param x: np.ndarray, shape = (?, ?)
        :return: x
        """
        return x

    def posthandle(self, y) -> object:
        """
        对y进行后处理
        原理:
        后处理只在LModel.predict函数中被调用
        当模型输出结果后:   :raw=False, label=False
        1. 先将y_pred(output)转化为y_pred(label)
        2. 然后对y_pred(label)进行后处理             <----------------------------------
        -> 最后返回'后处理'返回的结果

        *如果希望predict返回output, 那么调用函数时指定raw=True; (优先级更高)
        *如果希望predict返回label, 那么调用时指定参数label=True;
        :param y: object
        :return: y
        """
        return y

if __name__ == "__main__":
    class MyData(DataControl):
        def load(self):
            self.extend([1, 2, 3], [4, 5, 6])
            return True

        def labels2outputs(self, labels: Iterable) -> np.ndarray:
            return np.array(labels)

        def outputs2labels(self, outputs: np.ndarray) -> Iterable:
            return outputs


    class MyData2(MyData):
        ...

    pool = MyData()
    print(isinstance(MyData2, MyData))



# -*- coding: utf-8 -*-
from keras import layers as _L
from keras import Input
from Lutils._basic import *
import tensorflow as tf


# 定义Sigmoid类，继承自_L.Layer
class Sigmoid(_L.Layer):
    """
    Sigmoid(x) = 1 / (1 + exp(-x))，它的输出范围是(0, 1)，具有指数形状，可以表示为概率或者用于输入的归一化。它的缺点是容易出现饱和和梯度消失，以及输出不是零均值的。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(Sigmoid, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的sigmoid函数直接计算激活函数的值，并返回结果
        return tf.math.sigmoid(inputs)


# 定义Tanh类，继承自_L.Layer
class Tanh(_L.Layer):
    """
    Tanh(x) = (exp(x) - exp(-x)) / (exp(x) + exp(-x))，它的输出范围是(-1, 1)，是零均值的，比Sigmoid函数收敛速度更快。它的缺点也是容易出现饱和和梯度消失。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(Tanh, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的tanh函数直接计算激活函数的值，并返回结果
        return tf.math.tanh(inputs)


# 定义ReLU类，继承自_L.Layer
class ReLU(_L.Layer):
    """
    ReLU(x) = max(0, x)，它的输出范围是[0, +∞)，具有线性和非线性的特点，可以缓解梯度消失问题，训练速度快。它的缺点是当输入为负时，输出为零，导致部分神经元死亡，以及输出不是零均值的。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(ReLU, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的relu函数直接计算激活函数的值，并返回结果
        return tf.nn.relu(inputs)


# 定义LeakyReLU类，继承自_L.Layer
class LeakyReLU(_L.Layer):
    """
    Leaky ReLU(x) = max(αx, x)，其中α是一个很小的常数，例如0.01。它是为了解决ReLU函数在负区域导致神经元死亡的问题，让负区域有一个很小的梯度。它的优点是保留了ReLU函数的优势，缺点是输出仍然不是零均值的。
    """

    # 初始化方法，接收一个alpha参数，表示负区域的斜率系数，默认为0.01；和一个name参数，表示层的名称
    def __init__(self, alpha=0.01, name=None):
        # 调用父类的初始化方法
        super(LeakyReLU, self).__init__(name=name)
        # 保存alpha参数为一个属性
        self.alpha = alpha

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的leaky_relu函数直接计算激活函数的值，并返回结果
        return tf.nn.leaky_relu(inputs, alpha=self.alpha)


# 定义PReLU类，继承自_L.Layer
class PReLU(_L.Layer):
    """
    PReLU(x) = max(αx, x)，其中α是一个可学习的参数，可以通过反向传播进行更新。它是Leaky ReLU函数的一种改进，可以自适应地调整负区域的梯度。它在某些任务上表现得比ReLU函数更好，但也更容易过拟合。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(PReLU, self).__init__(name=name)
        # 定义一个可学习的参数alpha，初始值为0.25，范围为[0, 1]
        self.alpha = self.add_weight(name='alpha', shape=(), initializer='constant', trainable=True,
                                     constraint=keras.constraints.min_max_norm(min_value=0, max_value=1),
                                     dtype=tf.float32, default_value=0.25)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的maximum函数和multiply函数直接计算激活函数的值，并返回结果
        return tf.maximum(self.alpha * inputs, inputs)


# 定义ReLU6类，继承自_L.Layer
class ReLU6(_L.Layer):
    """
    ReLU6(x) = min(max(0, x), 6)，它是ReLU函数的一种变体，限制了输出范围在[0, 6]之间。它主要用于移动端设备上的低精度计算，以保证数值分辨率。它相比于ReLU函数有更好的饱和性，但也损失了一些表达能力。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(ReLU6, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的relu6函数直接计算激活函数的值，并返回结果
        return tf.nn.relu6(inputs)


# 定义Softsign类，继承自_L.Layer
class Softsign(_L.Layer):
    """
    Softsign(x) = x / (1 + |x|)，它是Tanh函数的一种替代选择，具有类似的曲线形状和输出范围。它相比于Tanh函数更平滑，导数下降更慢，可以缓解梯度消失问题。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(Softsign, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的softsign函数直接计算激活函数的值，并返回结果
        return tf.nn.softsign(inputs)


# 定义Softplus类，继承自_L.Layer
class Softplus(_L.Layer):
    """
    Softplus(x) = log(1 + exp(x))，它是ReLU函数的一种替代选择，具有类似的曲线形状和输出范围。它相比于ReLU函数更光滑，导数连续且非零，可以防止神经元死亡。它的缺点是不对称，不以零为中心，可能出现梯度消失的问题。
    """

    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(Softplus, self).__init__(name=name)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的softplus函数直接计算激活函数的值，并返回结果
        return tf.nn.softplus(inputs)


# 定义ELU类，继承自_L.Layer
class ELU(_L.Layer):
    """
    ELU(x) = x, if x > 0; α(exp(x) - 1), if x ≤ 0，其中α是一个常数，例如1。它是一种结合了Sigmoid和ReLU函数特点的激活函数，左侧具有软饱和性，右侧无饱和性。它可以缓解梯度消失问题，对输入变化或噪声更鲁棒，输出接近于零均值，收敛速度快。它的缺点是计算量相对较大。
    """
    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的elu函数直接计算激活函数的值，并返回结果
        return tf.nn.elu(inputs)


# 定义SELU类，继承自_L.Layer
class SELU(_L.Layer):
    """
    SELU(x) = λx, if x > 0; λα(exp(x) - 1), if x ≤ 0，其中λ和α是常数，例如1.0507和1.6733。它是ELU函数的一种改进，可以使得输入在经过多层之后保持一个固定的分布，从而提高网络的稳定性和收敛性。它的缺点是需要对输入进行标准化处理，以及对网络结构有一定的限制。
    """
    # 初始化方法，接收一个name参数，表示层的名称
    def __init__(self, name=None):
        # 调用父类的初始化方法
        super(SELU, self).__init__(name=name)
        # 定义两个常数lambda和alpha，用于计算激活函数的值
        self.lambd = 1.0507009873554804934193349852946
        self.alpha = 1.6732632423543772848170429916717

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的selu函数直接计算激活函数的值，并返回结果
        return tf.nn.selu(inputs)

# 定义SILU类，继承自_L.Layer
class SILU(_L.Layer):
    """
    SILU(x) = x * sigmoid(x)，其中sigmoid(x) = 1 / (1 + exp(-x))。它是一种结合了线性和非线性特性的激活函数，可以保留输入的信息，避免信息损失，同时也提高模型的表达能力。它可以实现自稳定，使得神经网络的隐藏层在训练过程中保持输出的均值和方差接近于1，从而提高网络的稳定性和收敛性。它可以近似任意凸函数，可以避免梯度消失和爆炸问题，可以提高神经网络的鲁棒性和容错性。
    """

    # 初始化函数
    def __init__(self):
        # 调用父类的初始化函数
        super(SILU, self).__init__()

    # 前向传播函数，输入为x
    def call(self, x):
        # 返回x乘以sigmoid(x)
        return x * tf.math.sigmoid(x)


# 定义GELU类，继承自_L.Layer
class GELU(_L.Layer):
    """
    GELU(x) = x * Φ(x)，其中Φ(x)是标准正态分布的累积分布函数。它是一种近似于ReLU函数的激活函数，具有平滑和非单调的特性，可以提高神经网络的表达能力和泛化能力
    """

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的gelu函数直接计算激活函数的值，并返回结果
        return tf.nn.gelu(inputs)


# 定义Swish类，继承自_L.Layer
class Swish(_L.Layer):
    """
    Swish(x) = x * sigmoid(βx)，其中β是一个可学习的参数或一个固定的常数。它是一种自门控的激活函数，可以根据输入自适应地调整信息流的量，可以提高神经网络的准确率和收敛速度
    """

    # 初始化方法，接收一个beta参数，表示sigmoid函数的系数，默认为1.0；和一个name参数，表示层的名称
    def __init__(self, beta=1.0, name=None):
        # 调用父类的初始化方法
        super(Swish, self).__init__(name=name)
        # 保存beta参数为一个属性
        self.beta = beta

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的sigmoid函数和multiply函数直接计算激活函数的值，并返回结果
        return tf.multiply(inputs, tf.math.sigmoid(self.beta * inputs))


# 定义Maxout类，继承自_L.Layer
class Maxout(_L.Layer):
    """
    Maxout(x) = max(w1x + b1, w2x + b2)，其中w1, w2, b1, b2是可学习的参数。它是一种分段线性的激活函数，可以近似任意凸函数，可以避免梯度消失和爆炸问题，可以提高神经网络的鲁棒性和容错性 。
    """

    # 初始化方法，接收一个units参数，表示输出张量的最后一维的大小；和一个name参数，表示层的名称
    def __init__(self, units, name=None):
        # 调用父类的初始化方法
        super(Maxout, self).__init__(name=name)
        # 保存units参数为一个属性
        self.units = units
        # 定义两个可学习的参数w1和w2，形状为(输入张量的最后一维大小, units)，初始值为随机正态分布
        self.w1 = self.add_weight(name='w1', shape=(None, units), initializer='random_normal', trainable=True)
        self.w2 = self.add_weight(name='w2', shape=(None, units), initializer='random_normal', trainable=True)
        # 定义两个可学习的参数b1和b2，形状为(units,)，初始值为0
        self.b1 = self.add_weight(name='b1', shape=(units,), initializer='zeros', trainable=True)
        self.b2 = self.add_weight(name='b2', shape=(units,), initializer='zeros', trainable=True)

    # 前向传播方法，接收一个inputs参数，表示输入张量
    def call(self, inputs):
        # 使用tensorflow的matmul函数和add函数计算两个线性变换的结果
        z1 = tf.add(tf.matmul(inputs, self.w1), self.b1)
        z2 = tf.add(tf.matmul(inputs, self.w2), self.b2)
        # 使用tensorflow的maximum函数直接计算激活函数的值，并返回结果
        return tf.maximum(z1, z2)


class Mish(_L.Layer):
    """
    Mish(x) = x * tanh(softplus(x))，其中softplus(x) = log(1 + exp(x))。它是一种自门控的激活函数，可以根据输入自适应地调整信息流的量，可以提高神经网络的准确率和收敛速度。它具有平滑和非单调的特性，可以提高神经网络的表达能力和泛化能力。它可以避免梯度消失和爆炸问题，可以提高神经网络的鲁棒性和容错性。
    """
    def __init__(self, name=None):
        super(Mish, self).__init__(name=name)

    def call(self, inputs):
        return inputs * tf.math.tanh(tf.math.softplus(inputs))


class Hardtanh(_L.Layer):
    """
    Hardtanh(x) = max(-1, min(1, x))，它是Tanh函数的一种变体，限制了输出范围在[-1, 1]之间。它主要用于低精度计算，以保证数值分辨率。它相比于Tanh函数有更好的饱和性，但也损失了一些表达能力。
    """
    def __init__(self, name=None):
        super(Hardtanh, self).__init__(name=name)

    def call(self, inputs):
        return tf.clip_by_value(inputs, -1, 1)


class ArcTan(_L.Layer):
    """
    ArcTan(x) = arctan(x)，它是一种反三角函数，具有类似于Tanh函数的曲线形状和输出范围。它相比于Tanh函数更平滑，导数下降更慢，可以缓解梯度消失问题。
    """
    def __init__(self, name=None):
        super(ArcTan, self).__init__(name=name)

    def call(self, inputs):
        return tf.math.atan(inputs)


# 定义RReLU类，继承自_L.Layer
class RReLU(_L.Layer):
    """
    RReLU是ReLU的一种变体，它在训练期间为负输入选择一个随机斜率，在测试期间使用平均斜率。这种方法有助于防止过拟合。
    """
    def __init__(self, alpha=0.1, name=None):
        super(RReLU, self).__init__(name=name)
        self.alpha = alpha

    def call(self, inputs):
        return tf.maximum(0., inputs) + self.alpha * tf.minimum(0., inputs)

# 定义Softmax类，继承自_L.Layer
class Softmax(_L.Layer):
    """
    Softmax函数用于多类分类问题，它将输出层的值通过激活函数映射到0-1区间，构造成概率分布。Softmax激活函数映射值越大，则真实类别可能性越大。
    """
    def call(self, inputs):
        return tf.nn.softmax(inputs)

# 定义Softmin类，继承自_L.Layer
class Softmin(_L.Layer):
    """
    Softmin函数是Softmax函数的一个变体，它对输入向量的每个元素应用负指数，然后对结果进行归一化。这样可以将任意长度的实向量压缩到0-1之间，并且向量中元素的总和为1。
    """
    def call(self, inputs):
        return tf.nn.softmin(inputs)

# 定义Softshrink类，继承自_L.Layer
class Softshrink(_L.Layer):
    """
    Softshrink是一种非线性激活函数，当输入值大于某个阈值时，输出为x-threshold；当输入值小于负阈值时，输出为x+threshold；否则输出为0。
    """
    def __init__(self, lambda_=0.5, name=None):
        super(Softshrink, self).__init__(name=name)
        self.lambda_ = lambda_

    def call(self, inputs):
        return tf.where(tf.abs(inputs) - self.lambda_ < 0, 0., inputs)

# 定义Hardshrink类，继承自_L.Layer
class Hardshrink(_L.Layer):
    """
    Hardshrink是一种非线性激活函数，它在输入值大于某个阈值时输出原始值，在输入值小于负阈值时也输出原始值，否则输出0。这种硬缩减操作可以帮助模型学习具有稀疏性的表示，有助于减少模型的计算复杂性和参数数量。
    """
    def __init__(self, lambda_=0.5, name=None):
        super(Hardshrink, self).__init__(name=name)
        self.lambda_ = lambda_

    def call(self, inputs):
        return tf.where(tf.abs(inputs) > self.lambda_, inputs, 0.)

# 定义Tanhshrink类，继承自_L.Layer
class Tanhshrink(_L.Layer):
    """
    Tanhshrink函数计算输入和tanh函数之间的差值，即x - tanh(x)。
    """
    def call(self, inputs):
        return inputs - tf.math.tanh(inputs)

# 定义Threshold类，继承自_L.Layer
class Threshold(_L.Layer):
    """
    Threshold（阈值）激活函数是一种最简单的非线性激活函数，它在神经网络中用于引入非线性性质。这种激活函数在输入大于某个阈值时输出1，否则输出0。简单来说，它模拟了一个开关的行为，输入超过阈值时激活（输出1），否则不激活（输出0）。
    """
    def __init__(self, threshold=1.0, value=0.0, name=None):
        super(Threshold, self).__init__(name=name)
        self.threshold = threshold
        self.value = value

    def call(self, inputs):
        return tf.where(inputs > self.threshold, inputs, self.value)

# 定义Hardswish类，继承自_L.Layer
class Hardswish(_L.Layer):
    """
    Hardswish是Swish激活函数的一种变体，它在输入值大于某个阈值或小于负阈值时输出原始值，否则输出的是输入值与（输入值+3）/6的乘积。这种硬编码的Swish只有在深层网络中才能发挥作。
    """
    def call(self, inputs):
        return inputs * tf.nn.relu6(inputs + 3.) / 6.

# 定义LogSigmoid类，继承自_L.Layer
class LogSigmoid(_L.Layer):
    """
    LogSigmoid激活函数是对Sigmoid函数求对数后得到的结果。它将输入的每个元素值x求以自然常数e为底的指数，然后再分别除以他们的和，最后取对数。
    """

    def call(self, inputs):
        return -tf.math.softplus(-inputs)

# 定义Hardsigmoid类，继承自_L.Layer
class Hardsigmoid(_L.Layer):
    """
    Hardsigmoid是一种近似于Sigmoid函数的非线性激活函数。它在输入值大于某个阈值或小于负阈值时输出原始值，否则输出的是输入值/6 + 1/2。
    """
    def call(self, inputs):
        return tf.nn.relu6(inputs + 3.) / 6.

# 定义Softmax2d类，继承自_L.Layer
class Softmax2d(_L.Layer):
    """
    Softmax2D激活函数是对二维矩阵（或张量）的操作，类似于一维数据中的Softmax激活函数，但是应用在矩阵的每一行上。这种操作常用于多类别分类任务，其中每一行代表一个样本的原始分数或logits，而每个列代表一个类别。
    """

    def call(self, inputs):
        return tf.nn.softmax(tf.reshape(inputs, (-1, inputs.shape[-1])))

# 定义LogSoftmax类，继承自_L.Layer
class LogSoftmax(_L.Layer):
    """
    LogSoftmax（对数软最大值）是一种常用的激活函数，通常用于多类别分类问题中神经网络的输出层。它将输入的原始分数（logits）转化为对数概率分布，以便更好地处理数值稳定性和训练的问题。
    """
    def call(self, inputs):
        return tf.math.log_softmax(inputs)

__act_cls = locals()


def doc(show_doc=True):
    # 罗列所有以_L.Layer作为父类的类对象
    _doc = ""
    for _cls in __act_cls:
        # 判断_cls是不是class
        if isinstance(__act_cls[_cls], type) and issubclass(__act_cls[_cls], _L.Layer):
            _doc += __act_cls[_cls].__name__ + ((":" + __act_cls[_cls].__doc__) if show_doc else "") + "\n\n"
    return _doc


if __name__ == "__main__":
    print(doc())



from lurara._basic import *
from lurara.act import *
from keras.layers import *
from keras import Model, Input
import tensorflow as tf


# Conv2D with BatchNormalization and SILU
class Conv2DBS:
    """
    (n, 1, 1, a) -> (n, 1 / strides, 1 / strides, filters)  # In Same Case
    基础的Conv2D with BatchNormalization and SILU结构
    BatchNormalization用于差异化明显的数据, 可降低这种差异带来的影响
    SILU用于非线性激活, 加速收敛, 且不改变输入的分布. 曲线和ReLU相似, 且比ReLU更快, 不会出现ReLU的梯度消失问题.
    """
    def __new__(cls, inputs, filters, kernel_size, strides, padding="same") -> tf.Tensor:
        x = Conv2D(filters, kernel_size, strides=strides, padding=padding, use_bias=False)(inputs)
        x = BatchNormalization()(x)
        x = SILU()(x)
        return x

    @staticmethod
    def TakeFeatures(inputs, filters) -> tf.Tensor:
        """
        (n, 1, 1, a) -> (n, .5, .5, filters)
        用于提取特征的ConvBS, 即
        ConvBS(inputs, filters, 3, 2)
        :param inputs:
        :param filters:
        :return:
        """
        x = Conv2DBS(inputs, filters, 3, 2)
        return x

    @staticmethod
    def ReChannels(inputs, filters) -> tf.Tensor:
        """
        (n, 1, 1, a) -> (n, 1, 1, filters)
        用于改变通道数的ConvBS, 即
        ConvBS(inputs, filters, 1, 1, 'valid')
        :param inputs:
        :param filters:
        :return:
        """
        x = Conv2DBS(inputs, filters, 1, 1, 'valid')
        return x

    @staticmethod
    def DarkBottle(inputs, filters, add=False) -> tf.Tensor:
        """
        (n, 1, 1, a) -> (n, 1, 1, filters)
        用于特征重整,并且可以缓减梯度消失问题
        ...
        :param inputs:
        :param filters:
        :param add: 前后add,缓减梯度消失. 但会影响神经元数据输出值
        :return:
        """
        x = Conv2DBS(inputs, filters // 2, 3, 1, padding='same')
        x = Conv2DBS(x, filters, 3, 1, padding='same')
        if add:
            return tf.add(x, inputs)
        return x

    @staticmethod
    def CSPConvB(inputs, filters, splits:int=2, add=False):
        """
        (n, 1, 1, a) -> (n, 1, 1, filters)
        用于特征重整和筛选,并且可以缓减梯度消失问题
        :param inputs:
        :param filters:
        :param splits:
        :param add: 前后add,缓减梯度消失. 但会影响神经元数据输出值
        :return:
        """
        assert isinstance(splits, int) and splits >= 1, 'splits must be integer and >= 1'
        x = Conv2DBS.ReChannels(inputs, filters)
        if splits == 1:
            xs = [x]
        else:
            xs = tf.split(x, splits, axis=-1)
        for i in range(splits):
            xs[i] = Conv2DBS.DarkBottle(xs[i], filters // splits, add=add)
        x = tf.concat([x] + xs, axis=-1)
        return Conv2DBS.ReChannels(x, filters)

    @staticmethod
    def CCUnit(inputs, filters, splits:int=2):
        """
        (n, 1, 1, a) -> (n, 1, 1, filters)
        用于特征重整和筛选,并且可以缓减梯度消失问题
        :param inputs:
        :param filters:
        :param splits:
        :return:
        """
        x = Conv2DBS.TakeFeatures(inputs, filters)
        x = Conv2DBS.CSPConvB(x, filters, splits=int(splits + .5), add=True)
        return x

# Conv1D with BatchNormalization and SILU
class Conv1DBS:
    """
    (n, 1, a) -> (n, 1 / strides, filters)  # In Same Case
    基础的Conv1D with BatchNormalization and SILU结构
    BatchNormalization用于差异化明显的数据, 可降低这种差异带来的影响
    SILU用于非线性激活, 加速收敛, 且不改变输入的分布. 曲线和ReLU相似, 且比ReLU更快, 不会出现ReLU的梯度消失问题.
    """
    def __new__(cls, inputs, filters, kernel_size, strides, padding="same") -> tf.Tensor:
        x = Conv1D(filters, kernel_size, strides=strides, padding=padding, use_bias=False)(inputs)
        x = BatchNormalization()(x)
        x = SILU()(x)
        return x

    @staticmethod
    def TakeFeatures(inputs, filters) -> tf.Tensor:
        """
        (n, 1, a) -> (n, .5, filters)
        用于提取特征的ConvBS, 即
        Conv1DBS(inputs, filters, 3, 2)
        :param inputs:
        :param filters:
        :return:
        """
        x = Conv1DBS(inputs, filters, 3, 2)
        return x

    @staticmethod
    def ReChannels(inputs, filters) -> tf.Tensor:
        """
        (n, 1, a) -> (n, 1, filters)
        用于改变通道数的Conv1DBS, 即
        Conv1DBS(inputs, filters, 1, 1, 'valid')
        :param inputs:
        :param filters:
        :return:
        """
        x = Conv1DBS(inputs, filters, 1, 1, 'valid')
        return x

    @staticmethod
    def DarkBottle(inputs, filters, add=False) -> tf.Tensor:
        """
        (n, 1, a) -> (n, 1, filters)
        用于特征重整,并且可以缓减梯度消失问题
        ...
        :param inputs:
        :param filters:
        :param add: 前后add,缓减梯度消失. 但会影响神经元数据输出值
        :return:
        """
        x = Conv1DBS(inputs, filters // 2, 3, 1, padding='same')
        x = Conv1DBS(x, filters, 3, 1, padding='same')
        if add:
            return tf.add(x, inputs)
        return x

    @staticmethod
    def CSPConvB(inputs, filters, splits:int=2, add=False):
        """
        (n, 1, a) -> (n, 1, filters)
        用于特征重整和筛选,并且可以缓减梯度消失问题
        :param inputs:
        :param filters:
        :param splits:
        :param add: 前后add,缓减梯度消失. 但会影响神经元数据输出值
        :return:
        """
        assert isinstance(splits, int) and splits > 1, 'splits must be integer and > 1'
        x = Conv1DBS.ReChannels(inputs, filters)
        xs = tf.split(x, splits, axis=-1)
        for i in range(splits):
            xs[i] = Conv1DBS.DarkBottle(xs[i], filters // splits, add=add)
        x = tf.concat([x] + xs, axis=-1)
        return Conv1DBS.ReChannels(x, filters)

    @staticmethod
    def CCUnit(inputs, filters, splits:int=2):
        """
        (n, 1, a) -> (n, 1, filters)
        用于特征重整和筛选,并且可以缓减梯度消失问题
        :param inputs:
        :param filters:
        :param splits:
        :return:
        """
        x = Conv1DBS.TakeFeatures(inputs, filters)
        x = Conv1DBS.CSPConvB(x, filters, splits=int(splits), add=True)
        return x


if  __name__ == "__main__":
    inputs = Input(shape=(224, 224, 3))
    x = Conv2DBS.TakeFeatures(inputs, 16)
    print(x)




import os

import numpy as np
import tensorflow as tf
from files3 import files
from typing import Tuple
from keras import layers as _L
from keras import Model, Input
from keras.callbacks import EarlyStopping
from keras.models import load_model
from lurara._basic import *
from lurara.data import DataControl


class LModel:
    """
    一个用于构建和管理机器学习模型的类。

    属性:
        patience: 一个整数，表示早停法的耐心值。
        data: 一个DataControl对象，用于控制数据的加载和处理。
        model: 一个tf.keras.Model对象，用于表示机器学习模型。
        optimizer: 一个tf.keras.optimizers对象，用于指定模型的优化器。
        loss: 一个tf.keras.losses对象或函数，用于指定模型的损失函数。
        metrics: 一个列表，包含tf.keras.metrics对象或函数，用于指定模型的评估指标。
        weighted_metrics: 一个列表，包含tf.keras.metrics对象或函数，用于指定模型的加权评估指标。
        callbacks: 一个列表，包含tf.keras.callbacks对象，用于指定模型的回调函数。
        batch_size: 一个整数，表示模型训练和评估时的批次大小。
        validation_split: 一个浮点数，表示从训练数据中划分出验证数据的比例。
    """

    patience = 10

    @property
    def trainable(self):
        """
        返回模型的训练状态。
        :return:
        """
        return self.model.trainable

    @trainable.setter
    def trainable(self, value):
        """
        对模型中的所有层设置是否可训练。
        :param value:
        :return:
        """
        for layer in self.model.layers:
            layer.trainable = value

    def __init__(self, input: Input, output:tf.Tensor, data_ctrl:DataControl=None, name=None):
        """
        初始化LModel类的实例。

        参数:
            input: 一个tf.keras.Input对象或列表，表示模型的输入层。
            output: 一个tf.Tensor对象或列表，表示模型的输出层。
            data_ctrl: 一个DataControl类或其子类，用于控制数据的加载和处理。
        异常:
            Exception: 如果data_ctrl不是DataControl类或其子类，或者无法创建DataControl实例，则抛出异常。
        """
        self.data:DataControl = data_ctrl
        if self.data is None:
            ...
        elif not isinstance(data_ctrl, DataControl):
            try:
                self.data = data_ctrl()
            except Exception as err:
                raise Exception(f"\n\nFailed to create DataControl instance. data_ctrl:{type(data_ctrl)}\n\terror info:\n\t{err}")

        # if input is None or output is None:
        #     raise ValueError(f"\n\ninput and output must be specified.\n\tinput:{input}\n\toutput:{output}")
        # else:
        #     self.model: Model = Model(inputs=input, outputs=output)
        self.model: Model = Model(inputs=input, outputs=output, name=name)

        # --------------------------------------------------------------------------------------------------------------
        self.optimizer = None
        self.loss = None
        self.metrics = None

        self.weighted_metrics = None
        self.callbacks = [EarlyStopping(patience=LModel.patience)]
        self.batch_size = None
        self.validation_split = 0.1

    def _check_model(self):
        assert self.model is not None, "Model is not defined. Please use .load(path) to load a model"
        assert isinstance(self.model, Model), f"Model is not a keras Model. Got {type(self.model)}"

    def _check_data(self):
        assert self.data is not None, "DataControl is not defined. Please use .set_data(DataControl) to set a data"
        assert isinstance(self.data, DataControl), f"DataControl is not a DataControl. Got {type(self.data)}"

    def config(self,
               optimizer=None,
               loss=None,
               metrics=None,
               weighted_metrics=None,
               callbacks=None,
               batch_size=None,
               validation_split=None):
        if optimizer is not None:
            self.optimizer = optimizer
        if loss is not None:
            self.loss = loss
        if metrics is not None:
            self.metrics = metrics
        if weighted_metrics is not None:
            self.weighted_metrics = weighted_metrics
        if callbacks is not None:
            _flag = False
            for each in callbacks:
                if isinstance(each, EarlyStopping):
                    _flag = True
                    break
            if not _flag:
                callbacks.append(EarlyStopping(patience=LModel.patience, restore_best_weights=True))
            self.callbacks = callbacks
        if batch_size is not None:
            self.batch_size = batch_size
        if validation_split is not None:
            if validation_split < 0.01:
                raise ValueError("validation_split should be >= 0.01")
            self.validation_split = validation_split

    def train(self, times:int=1, num=None, percent:float=1.0):
        """
        训练模型
        :param times: 训练的次数（轮数），默认为1
        :param num: 训练的数据量，如果指定了，就按照这个数量从数据集中随机抽取数据，否则按照percent参数来确定数据量
        :param percent: 训练的数据占总数据集的百分比，只有在num参数为空时才有效，默认为1.0（即使用全部数据）
        :return: 一个训练历史对象，包含了训练过程中的损失和评估指标等信息
        """
        self._check_model()
        self._check_data()
        self.model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics, weighted_metrics=self.weighted_metrics)

        _x, _y = self.data(num, percent)  # num优先级高于percent

        # print(f"len of x: {len(_x)}, len of y: {len(_y)}")
        _x = np.array(_x)
        _y = np.array(_y)

        his = self.model.fit(_x, _y, shuffle=True, batch_size=self.batch_size, epochs=times, validation_split=self.validation_split, callbacks=self.callbacks)
        return his

    def predict(self, input, pre=True, raw=False, label=False) -> object:
        """
       预测一个输入的输出结果
       :param input: 一个输入，可以是任意类型
       :param pre: 是否启用预处理，默认为True
       :param raw: 是否返回原始的输出结果，默认为False
       :param label: 是否返回输出结果的标签，默认为False
       :return: 一个输出结果，根据raw和label的设置，可能是一个数字、一个字符串或一个对象
       """
        self._check_model()
        self._check_data()
        if pre:
            input = self.data.prehandle(input)
        _result = self.model.predict(np.array(input).reshape((1, -1)))[0]
        if raw:
            return _result
        _result = self.data.output2label(_result)
        if label:
            return _result
        return self.data.posthandle(_result)

    def predicts(self, inputs: list, pre=True, raw=False, label=False) -> list:
        """
        预测一组输入的输出结果
        :param inputs: 一个输入的列表，每个输入可以是任意类型
        :param pre: 是否启用预处理，默认为True
        :param raw: 是否返回原始的输出结果，默认为False
        :param label: 是否返回输出结果的标签，默认为False
        :return: 一个输出结果的列表，根据raw和label的设置，每个输出结果可能是一个数字、一个字符串或一个对象
        """
        self._check_model()
        self._check_data()
        if pre:
            inputs = [self.data.prehandle(input) for input in inputs]
        _results = self.model.predict(np.array(inputs))
        if raw:
            return _results
        _results = [self.data.output2label(_result for _result in _results)]
        if label:
            return _results
        return [self.data.posthandle(_result) for _result in _results]

    def save(self, name):
        """
        保存模型权重到指定路径。

        参数:
            name: 表示保存模型的名称。

        异常:
            Exception: 如果没有创建模型，则抛出异常。
        """
        self._check_model()
        self.model.save_weights(name)


    def load(self, name):
        """
        从指定路径加载模型权重。
        * 注意: 若已创建模型，则会覆盖。

        参数:
            name: 表示保存模型的名称。

        异常:
            Exception: 如果没有创建模型，则抛出异常。
        """
        self.model.load_weights(name)
        self._check_model()

    def summary(self):
        """
        打印模型的摘要信息。
        """
        self._check_model()
        self.model.summary()

    def set_data(self, data):
        self.data = data
        self._check_data()

    def __str__(self):
        return self.model.name + "(LModel){" + f"inputs={self.model.input_shape}, outputs={self.model.output_shape}"  + "}"

    def __call__(self, inputs, *args, **kwargs):
        self._check_model()
        return self.model(inputs, *args, **kwargs)



from typing import Tuple
class Config:
    EXCLUDE_START = ['__', "_Config__", "EXCLUDE_START", "keys"]
    """
    Configure for function.
    Not save the key start_with '__'
    """
    def __init__(self, *a, **kwargs):
        self.__args = list(a)
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __iter__(self):
        return iter(self.__args)

    def __getitem__(self, item):
        return getattr(self, item)

    def keys(self):
        _ = []
        for k in dir(self):
            _flag = True
            for not_start in Config.EXCLUDE_START:
                if k.startswith(not_start):
                    _flag = False
                    break
            if _flag:
                _ += [k]
        return _

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(self.__args) + ", " + str({k: getattr(self, k) for k in self.keys()})


import os
import re

from files3 import files


# Standard Data Format
class StData:
    """
    定义了一种标准的数据集*.svoc, 构建了文件系统与标准数据集之间的桥梁. 只能在python代码中进行储存和读取
    xs, ys 或者 xs1, ys1  xs2, ys2 ... xsi, ysi ... xsn, ysn
    其中 xs, ys 均为list, x和y的类型没有明确限制,但x:object一般为ndarray, y:object一般为int|str
    上述两种格式均为支持的格式

    # -----------------------------------------------------------------

    一个数据集库只包含一种数据集格式, 即一个文件夹下只能包含一种数据集格式,最好不要将多个网络的数据集放在一个数据集库中管理
    每个数据集库按照index来管理, 比如len(stdata), stdata[i], 每个value为一个xs, ys

    """

    def __init__(self, fdir: str = ''):
        """

        :param fdir: 目标数据集所在的文件系统目录路径. 所有数据都将维护在此
        """
        assert isinstance(fdir, str), f"fdir must be str, not {type(fdir)}"
        fdir = re.sub('\s+$', '', fdir)
        fdir = re.sub('[\\/]+$', '', fdir)
        self.voc_f = files(fdir, '.svoc')
        self.voc_dir_name = os.path.dirname(fdir)

    def load(self, index):
        voc_list = self.voc_f.list()
        return self.voc_f[voc_list[index]]

    def save(self, xs, ys, mode: str = 'a', block_size: int = None):
        """
        保存数据到数据集库中
        :param xs:
        :param ys:
        :param mode: 'a' 'w' 两种, 'a'表示追加, 'w'表示覆盖(即会先清除所有原有数据集)
        :param block_size: None 用于将数据保存为多个部分, 每个部分的最大数据量. None表示保存到一个文件中
        :return:
        """
        mode = mode.lower()
        assert mode in ('a', 'w'), f"mode only accept 'a' or 'w', not {mode}"
        assert len(xs) == len(ys), f'xs and ys must have the same length, not {len(xs)} and {len(ys)}'
        if mode == 'w':
            del self.voc_f[:]

        # 计算要分成多少个block
        if block_size is None:
            block_num = 1
            block_size = len(xs)
        else:
            block_num = len(xs) // block_size
            if block_num * block_size < len(xs):
                block_num += 1

        ids = [-1] + [int(i) for i in self.voc_f.list()]
        id_max = max(ids) + 1

        for i in range(block_num):
            block_xs = xs[i * block_size: (i + 1) * block_size]
            block_ys = ys[i * block_size: (i + 1) * block_size]
            self.voc_f[str(i + id_max)] = block_xs, block_ys

    def __len__(self):
        return len(self.voc_f.list())

    def __getitem__(self, index):
        voc_list = self.voc_f.list()
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            xs, ys = [], []
            for i in range(start, stop, step):
                x, y = self.voc_f[voc_list[i]]
                xs.extend(x)
                ys.extend(y)
            return xs, ys
        else:
            return self.load(index)

    # readonly, no setitem


import numpy as np
import tensorflow as tf
from keras.layers import *
from keras import layers as _L, Sequential
from keras import Model, Input
from Lutils._basic import *
from Lutils.act import SILU
from Lutils.ops.conv import Conv2DBS, Conv1DBS
from Lutils.ops.sample import UCC2D, DCCC2D, UCC1D, DCCC1D



# 定义ConcatAny类
class ConcatA(Layer):
    """
    下面的输入都是可以被拼接起来的:
    shapes: (15,12) (15, 20) -> (15,20) (15, 20) -> concat
    shapes:  (12, 15) (10, 10) -> (12,15) (12, 15) -> concat
    """
    # 初始化函数
    def __init__(self, filters=None, padding="same", **kwargs):
        # 调用父类的初始化函数
        super(ConcatA, self).__init__(**kwargs)
        self.filters = filters
        self.padding = padding

    # 前向传播函数
    def call(self, list_inputs:list, axis=0, fill=0):
        # 获取输入的形状
        shapes = [tf.shape(input) for input in list_inputs]
        # 获取最大的形状
        max_shape = tf.reduce_max(tf.stack(shapes), axis=0)
        # 对每个输入进行填充
        padded_inputs = []
        for input in list_inputs:
            input_shape = tf.shape(input)
            paddings = tf.stack([tf.zeros_like(input_shape), max_shape - input_shape], axis=1)
            padded_input = tf.pad(input, paddings, constant_values=fill)
            padded_inputs.append(padded_input)
        # 拼接填充后的输入
        x = tf.concat(padded_inputs, axis=axis)

        if self.filters is None:
            return x

        x =  Conv2D(self.filters, 1, strides=1, padding=self.padding)(x)
        return x

# 定义类
class DimensionUp(Layer):
    # 初始化函数
    def __init__(self, input_shape=None, **kwargs):
        # 调用父类的初始化函数
        super(DimensionUp, self).__init__(**kwargs)
        self._input_shape = input_shape

    def call(self, inputs, *args, **kwargs):
        _new_shape = (*inputs.shape[1:], 1)
        return Reshape(_new_shape, input_shape=self._input_shape)(inputs)

class DimensionDown(Layer):
    # 初始化函数
    def __init__(self, input_shape=None, **kwargs):
        # 调用父类的初始化函数
        super(DimensionDown, self).__init__(**kwargs)
        self._input_shape = input_shape

    def call(self, inputs, *args, **kwargs):
        assert inputs.shape[-1] == 1, f"inputs.shape[-1] must be 1, but got {inputs.shape[-1]}"
        _new_shape = inputs.shape[1:-1]
        return Reshape(_new_shape, input_shape=self._input_shape)(inputs)

class WithDimensionUp(Layer):
    # 初始化函数
    def __init__(self, layer, input_shape=None, **kwargs):
        # 调用父类的初始化函数
        super(WithDimensionUp, self).__init__(**kwargs)
        self._input_shape = input_shape
        self.layer = layer

    def call(self, inputs, *args, **kwargs):
        _new_shape = (*inputs.shape[1:], 1)
        x = Reshape(_new_shape, input_shape=self._input_shape)(inputs)
        x = self.layer(x)
        assert x.shape[-1] == 1, f"x.shape[-1] must be 1, but got {inputs.shape[-1]}.This caused by your layer (which change the shape) to WithDimensionUp at initialize."
        _new_shape = x.shape[1:-1]
        return Reshape(_new_shape, input_shape=x.shape[1:])(x)

if  __name__ == '__main__':
    ...




from Lutils._basic import *
from Lutils.act import *
from keras.layers import *
from keras import Model, Input
import tensorflow as tf

from Lutils.ops.conv import Conv2DBS, Conv1DBS


class UCC2D:
    def __new__(cls, shallow_inputs, deep_inputs) -> tf.Tensor:
        """
        (n, 1, 1, a2), (n, 2, 2, a1) -> (n, 2, 2, a1)
        * a2 > a1 . gernaerally.
        :param shallow_inputs: (n, 2, 2, a1)
        :param deep_inputs: (n, 1, 1, a2)
        :return: (n, 2, 2, a1)
        """
        shallow_x = UpSampling2D()(deep_inputs)
        x = tf.concat([shallow_inputs, shallow_x], axis=-1)
        x = Conv2DBS.CSPConvB(x, shallow_inputs.shape[-1], add=False)
        return x


class DCCC2D:
    def __new__(cls, shallow_inputs, deep_inputs) -> tf.Tensor:
        """
        (n, 1, 1, a2), (n, 2, 2, a1) -> (n, 1, 1, a2)
        * a2 > a1 . gernaerally.
        :param shallow_inputs: (n, 2, 2, a1)
        :param deep_inputs: (n, 1, 1, a2)
        :return: (n, 2, 2, a1)
        """
        deep_x = Conv2DBS.TakeFeatures(shallow_inputs, deep_inputs.shape[-1])
        x = tf.concat([deep_inputs, deep_x], axis=-1)
        x = Conv2DBS.CSPConvB(x, deep_inputs.shape[-1], add=False)
        return x

class UCC1D:
    def __new__(cls, shallow_inputs, deep_inputs) -> tf.Tensor:
        """
        (n, 1, a2), (n, 2, a1) -> (n, 2, a1)
        * a2 > a1 . gernaerally.
        :param shallow_inputs: (n, 2, a1)
        :param deep_inputs: (n, 1, a2)
        :return: (n, 2, a1)
        """
        shallow_x = UpSampling1D()(deep_inputs)
        x = tf.concat([shallow_inputs, shallow_x], axis=-1)
        x = Conv1DBS.CSPConvB(x, shallow_inputs.shape[-1], add=False)
        return x


class DCCC1D:
    def __new__(cls, shallow_inputs, deep_inputs) -> tf.Tensor:
        """
        (n, 1, a2), (n, 2, a1) -> (n, 1, a2)
        * a2 > a1 . gernaerally.
        :param shallow_inputs: (n, 2, a1)
        :param deep_inputs: (n, 1, a2)
        :return: (n, 2, a1)
        """
        deep_x = Conv1DBS.TakeFeatures(shallow_inputs, deep_inputs.shape[-1])
        x = tf.concat([deep_inputs, deep_x], axis=-1)
        x = Conv1DBS.CSPConvB(x, deep_inputs.shape[-1], add=False)
        return x




import cv2
import numpy as np


def mosaic_augmentation(xs, ys, img_size=640, **k):
    """
    Perform mosaic data augmentation on a dataset.
    :param xs: list of images
    :param ys: list of bounding boxes [cx, cy, w, h, cls] or list of classes
    :param img_size: size of the output image
    :return: list of augmented images and list of new bounding boxes or classes
    """
    # List to store new images and bounding boxes or classes
    new_xs = []
    new_ys = []

    for i in range(0, len(xs), 4):
        # Create an empty image of the specified size
        output_image = np.zeros((img_size, img_size, 3))

        for j in range(4):
            # Choose an image
            index = i + j if i + j < len(xs) else i + j - len(xs)
            img = xs[index]
            target = ys[index]

            # Resize the image to the desired size
            h, w, _ = img.shape
            if w > h:
                img = cv2.resize(img, (img_size, int(img_size * h / w)))
            else:
                img = cv2.resize(img, (int(img_size * w / h), img_size))

            # Determine where to place the image based on its quadrant
            if j == 0:  # top-left
                output_image[:img.shape[0], :img.shape[1]] = img
                if isinstance(target[0], list):  # ys is a list of bounding boxes
                    new_ys.append([target[0] * img.shape[1], target[1] * img.shape[0], target[2] * img.shape[1],
                                   target[3] * img.shape[0], target[4]])
                else:  # ys is a list of classes
                    new_ys.append(target)
            elif j == 1:  # top-right
                output_image[:img.shape[0], -img.shape[1]:] = img
                if isinstance(target[0], list):  # ys is a list of bounding boxes
                    new_ys.append(
                        [img_size - target[0] * img.shape[1], target[1] * img.shape[0], target[2] * img.shape[1],
                         target[3] * img.shape[0], target[4]])
                else:  # ys is a list of classes
                    new_ys.append(target)
            elif j == 2:  # bottom-left
                output_image[-img.shape[0]:, :img.shape[1]] = img
                if isinstance(target[0], list):  # ys is a list of bounding boxes
                    new_ys.append(
                        [target[0] * img.shape[1], img_size - target[1] * img.shape[0], target[2] * img.shape[1],
                         target[3] * img.shape[0], target[4]])
                else:  # ys is a list of classes
                    new_ys.append(target)
            elif j == 3:  # bottom-right
                output_image[-img.shape[0]:, -img.shape[1]:] = img
                if isinstance(target[0], list):  # ys is a list of bounding boxes
                    new_ys.append([img_size - target[0] * img.shape[1], img_size - target[1] * img.shape[0],
                                   target[2] * img.shape[1], target[3] * img.shape[0], target[4]])
                else:  # ys is a list of classes
                    new_ys.append(target)

        new_xs.append(output_image)

    return new_xs, new_ys


def random_affine(xs, ys, degrees=10, translate=0.1, scale=0.1, shear=0.1, **k):
    new_xs = []
    new_ys = []

    for x, y in zip(xs, ys):
        # 获取图像的宽度和高度
        h, w = x.shape[:2]
        # 创建仿射变换矩阵
        angle = np.random.uniform(-degrees, degrees)
        tx = np.random.uniform(-translate, translate) * w
        ty = np.random.uniform(-translate, translate) * h
        s = np.random.uniform(1 - scale, 1 + scale)
        shear = np.random.uniform(-shear, shear)
        M = cv2.getRotationMatrix2D((w / 2, h / 2), angle, s)
        M[0, 2] += tx
        M[1, 2] += ty
        M[1, 0] += shear
        # 对图像进行仿射变换
        new_x = cv2.warpAffine(x, M, (w, h), flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_REFLECT)
        new_xs.append(new_x)

        if isinstance(y, (list, np.ndarray)):
            new_y = []
            for bbox in y:
                cx, cy, w, h, cls = bbox
                cx *= x.shape[1]
                cy *= x.shape[0]
                w *= x.shape[1]
                h *= x.shape[0]

                # 将bbox的中心点坐标和宽高转换为左上角和右下角的坐标
                x1, y1, x2, y2 = cx - w / 2, cy - h / 2, cx + w / 2, cy + h / 2
                # 创建表示bbox的3x3的坐标矩阵
                coords = np.array([[x1, y1, 1], [x2, y1, 1], [x2, y2, 1], [x1, y2, 1]])
                # 对坐标矩阵进行仿射变换
                new_coords = np.dot(coords, M.T)
                # 计算新的bbox的中心点坐标和宽高
                new_x1, new_y1 = new_coords.min(axis=0)
                new_x2, new_y2 = new_coords.max(axis=0)
                new_cx, new_cy = (new_x1 + new_x2) / 2, (new_y1 + new_y2) / 2
                new_w, new_h = new_x2 - new_x1, new_y2 - new_y1
                # 添加新的bbox到列表中
                new_y.append(
                    [new_cx / new_x.shape[1], new_cy / new_x.shape[0], new_w / new_x.shape[1], new_h / new_x.shape[0], cls])
            new_ys.append(new_y)
        else:
            new_ys = ys

    return new_xs, new_ys


def augment_hsv(xs, ys, hgain=0.1, sgain=0.1, vgain=0.1, **k):
    """
    Perform HSV data augmentation on a dataset.
    :param xs: list of images
    :param ys: list of bounding boxes [cx, cy, w, h, cls] or list of classes
    :param hgain: hue gain
    :param sgain: saturation gain
    :param vgain: value gain
    :return: list of augmented images and list of new bounding boxes or classes
    """
    # List to store new images and bounding boxes or classes
    new_xs = []
    new_ys = []

    for img, target in zip(xs, ys):
        # Convert image to HSV color space
        img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

        # Apply transformations
        hue = np.random.uniform(1 - hgain, 1 + hgain)
        sat = np.random.uniform(1 - sgain, 1 + sgain)
        val = np.random.uniform(1 - vgain, 1 + vgain)
        img_hsv = (img_hsv * np.array([hue, sat, val])).clip(min=0, max=255).astype(np.uint8)

        # Convert image back to BGR color space
        img = cv2.cvtColor(img_hsv, cv2.COLOR_HSV2BGR)

        new_xs.append(img)
        new_ys.append(target)

    return new_xs, new_ys


def horizontal_flip(xs, ys, **k):
    """
    Perform Horizontal Flip data augmentation on a dataset.
    :param xs: list of images
    :param ys: list of bounding boxes [cx, cy, w, h, cls] or list of classes
    :return: list of augmented images and list of new bounding boxes or classes
    """
    # List to store new images and bounding boxes or classes
    new_xs = []
    new_ys = []

    for img, target in zip(xs, ys):
        # Apply horizontal flip
        img = cv2.flip(img, 1)

        if isinstance(target[0], list):  # ys is a list of bounding boxes
            # Flip bounding boxes
            target = [[1 - box[0], box[1], box[2], box[3], box[4]] for box in target]

        new_xs.append(img)
        new_ys.append(target)

    return new_xs, new_ys


def voc2clsfy(xs, ys, **kwargs):
    new_xs = []
    new_ys = []

    for x, y in zip(xs, ys):
        # 获取图像的宽度和高度
        h, w = x.shape[:2]

        for bbox in y:
            cx, cy, bw, bh, cls = bbox
            # 计算bbox的左上角和右下角的坐标
            x1 = int((cx - bw/2) * w)
            y1 = int((cy - bh/2) * h)
            x2 = int((cx + bw/2) * w)
            y2 = int((cy + bh/2) * h)
            # 创建一个新的随机子图
            new_x = np.random.randint(0, 255, (h, w, 3), dtype=np.uint8)
            # 将原图中对应区域的图片拷贝到新子图中
            new_x[y1:y2, x1:x2] = x[y1:y2, x1:x2]
            # 添加新的子图和类别到列表中
            new_xs.append(new_x)
            new_ys.append(cls)

    return new_xs, new_ys




import os
import glob
import cv2
import numpy as np
import xml.etree.ElementTree as ET
from keras import Input
from keras.layers import Conv2D, Dense, Concatenate, Flatten
from lurara.main import LModel
from lurara.act import Softmax, SILU
from lurara.ops.conv import Conv2DBS
from lurara.ops.pooling import SPPF2D
from lurara.ops.sample import UCC2D, DCCC2D


def draw_label(img, bboxes):
    """
    绘制检测框
    :param img: 输入图像
    :param bboxes: list of [cx, cy, w, h, cls]，其中cx, cy, w, h均为float,取值0-1
    :return: img
    """
    img_h, img_w = img.shape[:2]

    for bbox in bboxes:
        cx, cy, w, h, cls = bbox
        x_min = int((cx - w / 2) * img_w)
        y_min = int((cy - h / 2) * img_h)
        x_max = int((cx + w / 2) * img_w)
        y_max = int((cy + h / 2) * img_h)

        # 生成随机颜色
        color = np.random.randint(0, 255, 3)
        color_index = np.random.choice(3)
        color[color_index] //= 2
        color[color_index] += 127
        for i in range(3):
            if i == color_index:
                continue
            color[i] /= 2
        color_txt = color.copy()
        color_txt[color_index] *= 0.75
        for i in range(3):
            if i == color_index:
                continue
            color_txt[i] *= 1.25

        # 生成文本颜色
        color_txt = color_txt.astype(int)

        cv2.rectangle(img, (x_min, y_min), (x_max, y_max), tuple(color.tolist()), 2)
        cv2.putText(img, str(cls), (x_min, y_min - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.9, tuple(color_txt.tolist()), 2)

    return img


def load_pascal_voc_dataset(jpeg_path, anno_path, cls_txt_path):
    # 获取所有JPEG图像的路径
    img_paths = glob.glob(os.path.join(jpeg_path, '*.jpg'))

    # 读取类别文件
    with open(cls_txt_path, 'r', encoding='utf-8') as f:
        classes = f.read().splitlines()

    bboxes = []
    for img_path in img_paths:
        # 获取对应的注释文件路径
        base_name = os.path.basename(img_path).replace('.jpg', '.xml')
        anno_file = os.path.join(anno_path, base_name)

        # 判断文件是否存在
        if not os.path.exists(anno_file):
            print("[Warning]: Can not find jpeg's annotation file: {}".format(anno_file))
            bboxes.append([])
            continue


        # 解析XML文件
        tree = ET.parse(anno_file)
        root = tree.getroot()

        # 获取图像的宽度和高度
        size = root.find('size')
        img_width = float(size.find('width').text)
        img_height = float(size.find('height').text)

        img_bboxes = []
        for obj in root.findall('object'):
            cls = classes.index(obj.find('name').text)

            bbox = obj.find('bndbox')
            xmin = float(bbox.find('xmin').text) / img_width
            ymin = float(bbox.find('ymin').text) / img_height
            xmax = float(bbox.find('xmax').text) / img_width
            ymax = float(bbox.find('ymax').text) / img_height

            # 计算中心点坐标和宽度、高度
            cx = (xmin + xmax) / 2
            cy = (ymin + ymax) / 2
            w = xmax - xmin
            h = ymax - ymin

            img_bboxes.append([cx, cy, w, h, cls])

        bboxes.append(np.array(img_bboxes))

    return img_paths, bboxes, classes


def build_yolov8(inputs_shape: tuple, classes: list, model_size: str = 'n', reg_max=4, debug=False) -> dict:
    """
    Create YoloV8 Net
    *return a dict with keys: body neck head
    :param inputs_shape:
    :param classes: list of str
    :param model_size: str, n s m l x
    :return: dict with keys: body neck head
    """
    d, w, r = -1, -1, -1  # deepen_factor, widden_factor, ratio
    model_size = model_size.lower()
    if model_size == 'n':
        d, w, r = 0.67, 0.25, 2.0
    elif model_size == 's':
        d, w, r = 0.67, 0.50, 2.0
    elif model_size == 'm':
        d, w, r = 1.0, 0.75, 1.5
    elif model_size == 'l':
        d, w, r = 1.34, 1.00, 1.0
    elif model_size == 'x':
        d, w, r = 1.34, 1.25, 1.0
    else:
        raise ValueError('Unknown model size: {}'.format(model_size))

    # build_body:
    body_inputs = Input(inputs_shape)  # 参考shape: 640 640 3
    x = Conv2DBS.TakeFeatures(body_inputs, filters=int(64 * w + .5))  # -> n 320 320 64*w
    # Stage1
    x1 = Conv2DBS.CCUnit(x, filters=int(128 * w + .5), splits=3 * d)  # x1 -> n 160 160 128*w
    # Stage2
    x2 = Conv2DBS.CCUnit(x1, filters=int(256 * w + .5), splits=3 * d)  # x2 -> n 80 80 256*w
    # Stage3
    x3 = Conv2DBS.CCUnit(x2, filters=int(512 * w + .5), splits=3 * d)  # x3 -> n 40 40 512*w
    # Stage4
    x = Conv2DBS.CCUnit(x3, filters=int(512 * w * r + .5), splits=3 * d)  # x4 -> n 20 20 512*w*r
    x4 = SPPF2D(x, int(512 * w * r + .5))

    net_body = LModel(input=body_inputs, output=[x2, x3, x4], name='YOLOv8_CSPDarknet')
    if debug: print("Create Body Finish!")
    # ------------------------------------

    cls_inputs = Input(inputs_shape)
    xs = [Flatten()(xsi) for xsi in net_body(cls_inputs)]
    x = Concatenate()(xs)
    x = Dense(len(classes))(x)
    x = Softmax()(x)
    net_body_cls = LModel(input=cls_inputs, output=x, name='YOLOv8_Pretrain_Classifier')
    if debug: print("Create Body Classifier Finish!")
    # ------------------------------------

    u1 = UCC2D(x3, x4)  # u1 -> x3 -> n 40 40 512*w
    u2 = UCC2D(x2, u1)  # u2 -> x2 -> n 80 80 256*w

    d1 = DCCC2D(u2, u1)  # d1 -> u1 -> n 40 40 512*w
    d2 = DCCC2D(d1, x4)  # d2 -> x4 -> n 20 20 512*w*r

    net_neck = LModel(input=body_inputs, output=[u2, d1, d2],  name='YOLOv8_CSPDarknet_PAFPN')
    cls_inputs = Input(inputs_shape)
    xs = net_neck(cls_inputs)
    xs = [Flatten()(xs_i) for xs_i in xs]
    x = Concatenate()(xs)
    x = Dense(len(classes))(x)
    x = Softmax()(x)
    net_neck_cls = LModel(cls_inputs, x,  name='YOLOv8_Neck_Pretrain_Classifier')
    if debug: print("Create Neck Finish!")
    # ------------------------------------

    h1_box = Conv2DBS(u2, 3, 1, 1)
    h1_box = Conv2D(4 * reg_max, 1, 1, 'valid')(h1_box)
    h1_cls = Conv2DBS(u2, 3, 1, 1)
    h1_cls = Conv2D(len(classes), 1, 1, 'valid')(h1_cls)

    h2_box = Conv2DBS(d1, 3, 1, 1)
    h2_box = Conv2D(4 * reg_max, 1, 1, 'valid')(h2_box)
    h2_cls = Conv2DBS(d1, 3, 1, 1)
    h2_cls = Conv2D(len(classes), 1, 1, 'valid')(h2_cls)

    h3_box = Conv2DBS(d2, 3, 1, 1)
    h3_box = Conv2D(4 * reg_max, 1, 1, 'valid')(h3_box)
    h3_cls = Conv2DBS(d2, 3, 1, 1)
    h3_cls = Conv2D(len(classes), 1, 1, 'valid')(h3_cls)

    net_head = LModel(input=body_inputs, output=[h1_box, h1_cls, h2_box, h2_cls, h3_box, h3_cls], name='YOLOv8')
    if debug: print("Create Head Finish!")
    # ------------------------------------

    return {
        "body": net_body,
        "neck": net_neck,
        "head": net_head,
        "yolo": net_head,
        "body_cls": net_body_cls,
        "neck_cls": net_neck_cls
    }


if __name__ == "__main__":
    _ = load_pascal_voc_dataset(r'D:\desktop20231030\yolov8\VOC_UG\JPEGImages', r'D:\desktop20231030\yolov8\VOC_UG\Anotations', r'D:\desktop20231030\yolov8\VOC_UG\predefined_classes.txt')
    print(*_, sep='\n')



from Lutils.ops.conv import Conv2DBS, Conv1DBS
from keras.layers import Concatenate, MaxPool2D

class SPPF2D:
    def __new__(cls, inputs, filters):
        x0 = Conv2DBS.ReChannels(inputs, filters)
        x1 = MaxPool2D(2, 1, 'same')(x0)
        x2 = MaxPool2D(3, 1, 'same')(x1)
        x3 = MaxPool2D(4, 1, 'same')(x2)
        x = Concatenate(axis=-1)([x0, x1, x2, x3])
        x = Conv2DBS.ReChannels(x, filters)
        return x


class SPPF1D:
    def __new__(cls, inputs, filters):
        x0 = Conv1DBS.ReChannels(inputs, filters)
        x1 = MaxPool1D(2, 1, 'same')(x0)
        x2 = MaxPool1D(3, 1, 'same')(x1)
        x3 = MaxPool1D(4, 1, 'same')(x2)
        x = Concatenate(axis=-1)([x0, x1, x2, x3])
        x = Conv1DBS.ReChannels(x, filters)
        return x


#!/usr/bin/env python
# coding:utf-8

from setuptools import find_packages, setup

setup(
    name='lurara',
    version='0.1.2',
    description='more easy use for keras(care for it used in small or mid model, not efficiency). simplify like dataset create, model train&predict. Used by person',
    author_email='2229066748@qq.com',
    maintainer="Eagle'sBaby",
    maintainer_email='2229066748@qq.com',
    packages=find_packages(),
    license='Apache Licence 2.0',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    keywords = ['keras', 'utils'],
    python_requires='>=3',
    install_requires=[
        "tensorflow==2.6.0",
        "keras==2.6.0",
        "files3"
    ],
)




import cv2
import numpy as np
from copy import deepcopy

def fillresize(x:np.ndarray, new_size:tuple, fill:float) -> np.ndarray:
    # Compute the ratio of the new size and the original size
    ratio = min(new_size[0] / x.shape[0], new_size[1] / x.shape[1])

    # Compute the size of the resized image
    resize_size = (round(x.shape[1] * ratio), round(x.shape[0] * ratio))

    # Resize the image
    x_resized = cv2.resize(x, resize_size, interpolation = cv2.INTER_AREA)

    # Create a new array filled with the fill value
    new_x = np.full((*new_size, x.shape[2]), fill, dtype=x.dtype)

    # Compute the starting indices for x_resized
    start_y = (new_size[0] - x_resized.shape[0]) // 2
    start_x = (new_size[1] - x_resized.shape[1]) // 2

    # Copy x_resized into the center of new_x
    new_x[start_y:start_y+x_resized.shape[0], start_x:start_x+x_resized.shape[1]] = x_resized

    return new_x

def bboxresize(y: list, old_size: tuple, new_size: tuple) -> list:
    """
    Resize bounding boxes according to the ratio of the new size and the original size
    :param y: list of [cx, cy, w, h, cls], and cx cy w h is float 0 - 1
    :param old_size:
    :param new_size:
    :return:
    """
    y = deepcopy(y)
    # Compute the ratio of the new size and the original size
    _y, _x = new_size[0] / old_size[0], new_size[1] / old_size[1]
    if _y == _x:
        return y

    if 1 < _y < _x:
        _ratio = [_y / _x, 1]
    elif _y < _x < 1:
        _ratio = [1, _y / _x]

    if 1 < _x < _y:
        _ratio = [1, _x / _y]
    elif _x < _y < 1:
        _ratio = [_x / _y, 1]

    for bbox in y:
        if _ratio[1] != 1:
            bbox[0] = (bbox[0] - 0.5) * _ratio[1] + .5
        if _ratio[0] != 1:
            bbox[1] = (bbox[1] - 0.5) * _ratio[0] + 0.5
        bbox[2] *= _ratio[1]
        bbox[3] *= _ratio[0]

    return y



import concurrent.futures
import random

from lurara.augment.img_augment import *


class AugmentPipe:
    """
    用户将需要使用的数据增强的函数依次加入pipe, 这些函数必须以xs, ys作为通用的形参, 并且使用通用的kwargs
    """

    def __init__(self, use_multithreading: bool = False, num_threads: int = 4):
        self.augmentations = []
        self.use_multithreading = use_multithreading
        self.num_threads = num_threads
        if use_multithreading:
            assert num_threads > 1, f"num_threads must be greater than 1, but got {num_threads}"

    def add(self, func, prob=1.0):
        self.augmentations.append((func, prob))

    def remove(self, func):
        self.augmentations = [(f, p) for f, p in self.augmentations if f != func]

    def __call__(self, xs, ys, **kwargs):
        if self.use_multithreading:
            assert len(xs) == len(
                ys), f"len(xs) must be equal to len(ys), but got len_xs:{len(xs)} and len_xs:{len(ys)}"
            assert len(
                xs) >= self.num_threads, f"len(xs) must be greater than num_threads, but got len_xs:{len(xs)} and num_threads:{self.num_threads}"
            # Calculate the size of each chunk
            chunk_size = len(xs) // self.num_threads

            # Create a ThreadPoolExecutor
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.num_threads - 1) as executor:
                # Create futures for each thread
                futures = [executor.submit(self._process, xs[i * chunk_size:(i + 1) * chunk_size],
                                           ys[i * chunk_size:(i + 1) * chunk_size],
                                           i * chunk_size, (i + 1) * chunk_size, **kwargs) for i in
                           range(self.num_threads - 1)]
                # Process one part of the data in the main thread
                _start = chunk_size * (self.num_threads - 1)
                self._process(xs[_start:], ys[_start:], 0, -1, chunk_size, **kwargs)
                # Wait for all futures to complete
                concurrent.futures.wait(futures)
                # Get the results
                results = [f.result() for f in futures]
                new_xs = [img for sublist in results for img, _ in sublist]
                new_ys = [target for sublist in results for _, target in sublist]
        else:
            new_xs, new_ys = self._process(xs, ys, 0, -1, **kwargs)
        return new_xs, new_ys

    def _process(self, xs, ys, start, end, **kwargs):
        new_xs = xs.copy()
        new_ys = ys.copy()
        for func, prob in self.augmentations:
            if random.random() < prob:
                try:
                    new_xs, new_ys = func(new_xs, new_ys, **kwargs)
                except Exception as e:
                    raise Exception(f"Error occurred in function {func.__name__} for indexes[{start}, {end}): {str(e)}")
        return new_xs, new_ys


if __name__ == "__main__":
    from lurara.prefab.yolo import *

    _save_dir = r"C:\Users\22290\Desktop\yolovoc"
    _voc_path = r"D:\desktop20231030\yolov8\VOC_UG"
    _cls_txt_name = "predefined_classes.txt"
    img_paths, bboxes = load_pascal_voc_dataset(os.path.join(_voc_path, "JPEGImages"),
                                                os.path.join(_voc_path, "Annotations"),
                                                os.path.join(_voc_path, _cls_txt_name))
    imgs = [cv2.imread(img_path) for img_path in img_paths]

    # degrees=10, translate=.1, scale=.1, shear=10
    config = {
        'degrees': 10,
        'translate': 0.1,
        'scale': 0.1,
    }

    ap = AugmentPipe()

    ap.add(random_affine, prob=1)
    ap.add(augment_hsv, prob=1)
    ap.add(horizontal_flip, prob=0.5)

    new_imgs, new_bboxes = ap(imgs, bboxes, **config)
    # new_imgs, new_bboxes = imgs, bboxes
    # draw rect on img:
    for i, (img, bboxes) in enumerate(zip(new_imgs, new_bboxes)):
        h, w = img.shape[:2]
        for bbox in bboxes:
        #     # bbox is cx, cy, w, h
        #     # trans float to int by shape
            cx, cy, bw, bh = [int(bbox[0] * w), int(bbox[1] * h), int(bbox[2] * w), int(bbox[3] * h)]
            cv2.rectangle(img, (cx - bw // 2, cy - bh // 2), (cx + bw // 2, cy + bh // 2), (0, 0, 255), 2)
        cv2.imwrite(os.path.join(_save_dir, f"augmented_{i}.jpg"), img)